__version__ = "0.1.38"
from ._ledger import (
    GeneralLedgerLine,
    GeneralLedgerLineImpl,
    LedgerElement,
    LedgerElementImpl,
    LedgerLine,
    LedgerLineImpl,
    MultiLedgerLine,
    MultiLedgerLineImpl,
    generalledger_line_to_multiledger_line,
    generalledger_to_multiledger,
    multiledger_line_to_generalledger_line,
    multiledger_line_to_ledger_line,
    multiledger_to_generalledger,
    multiledger_to_ledger,
)
from ._multidimensional import multidimensional_ledger_to_ledger
from ._sheets import get_sheets
from .reader import (
    ledger_from_expenses,
    ledger_from_sales,
    read_all_csvs,
    read_general_ledger,
    read_simple_csvs,
    withholding_tax,
)

__all__ = [
    "GeneralLedgerLine",
    "GeneralLedgerLineImpl",
    "LedgerElement",
    "LedgerElementImpl",
    "LedgerLine",
    "LedgerLineImpl",
    "MultiLedgerLine",
    "MultiLedgerLineImpl",
    "generalledger_line_to_multiledger_line",
    "generalledger_to_multiledger",
    "get_sheets",
    "ledger_from_expenses",
    "ledger_from_sales",
    "multidimensional_ledger_to_ledger",
    "multiledger_line_to_generalledger_line",
    "multiledger_line_to_ledger_line",
    "multiledger_to_generalledger",
    "multiledger_to_ledger",
    "read_all_csvs",
    "read_general_ledger",
    "read_simple_csvs",
    "withholding_tax",
]



from collections.abc import Sequence
from decimal import Decimal
from itertools import chain
from typing import Any, Callable, Literal, Protocol, TypeVar

import attrs
import pandas as pd
from account_codes_jp._common import SUNDRY

AccountSundry = Literal["諸口"]
Account = TypeVar("Account", bound=str)
Currency = TypeVar("Currency", bound=str)


class _LedgerLineBase(Protocol):
    date: pd.Timestamp
    """The date when the transaction occurred."""


class LedgerLine(_LedgerLineBase, Protocol[Account, Currency]):
    amount: Decimal
    """The amount. Must be non-negative."""
    currency: Currency
    """The currency."""
    debit_account: Account
    """The account written on the debit side."""
    credit_account: Account
    """The account written on the credit side."""


class LedgerElement(Protocol[Account, Currency]):
    account: Account
    """The account."""
    amount: Decimal
    """The amount."""
    currency: Currency
    """The currency."""


class MultiLedgerLine(_LedgerLineBase, Protocol[Account, Currency]):
    debit: Sequence[LedgerElement[Account, Currency]]
    """The accounts and amounts on the debit side.
    Each amount needs to be non-negative."""
    credit: Sequence[LedgerElement[Account, Currency]]
    """The accounts and amounts on the credit side.
    Each amount needs to be non-negative."""


class GeneralLedgerLine(_LedgerLineBase, Protocol[Account, Currency]):
    values: Sequence[LedgerElement[Account, Currency]]
    """The accounts and amounts. Amounts does not need to be non-negative."""


@attrs.frozen(kw_only=True)
class LedgerLineImpl(LedgerLine[Account, Currency]):
    date: pd.Timestamp
    amount: Decimal = attrs.field()
    currency: Currency
    debit_account: Account
    credit_account: Account

    def __repr__(self) -> str:
        return (
            f"{self.date} {self.amount} {self.currency} "
            f"{self.debit_account} / {self.credit_account}"
        )

    @amount.validator
    def _validate_amount(self, attribute: Any, value: Decimal) -> None:
        if value < Decimal(0):
            raise ValueError("amount must be non-negative")


@attrs.frozen(kw_only=True)
class LedgerElementImpl(LedgerElement[Account, Currency]):
    account: Account
    amount: Decimal
    currency: Currency

    def __repr__(self) -> str:
        return f"{self.account} {self.amount} {self.currency}"


@attrs.frozen(kw_only=True, auto_detect=True)
class MultiLedgerLineImpl(MultiLedgerLine[Account, Currency]):
    date: pd.Timestamp
    debit: Sequence[LedgerElement[Account, Currency]] = attrs.field()
    credit: Sequence[LedgerElement[Account, Currency]] = attrs.field()

    def __repr__(self) -> str:
        date = pd.Series([self.date], name="date")
        debit = pd.DataFrame(
            self.debit, columns=["debit_account", "amount", "currency"]
        )
        credit = pd.DataFrame(
            self.credit, columns=["credit_account", "amount", "currency"]
        )
        return (
            pd.concat([date, debit, credit], axis=1)
            .replace({pd.NaT: ""})
            .fillna("")
            .to_string(index=False, header=False)
        )

    @debit.validator
    def _validate_debit(
        self, attribute: Any, value: Sequence[LedgerElement[Account, Currency]]
    ) -> None:
        if any(el.amount < Decimal(0) for el in value):
            raise ValueError("amount must be non-negative")

    @credit.validator
    def _validate_credit(
        self, attribute: Any, value: Sequence[LedgerElement[Account, Currency]]
    ) -> None:
        if any(el.amount < Decimal(0) for el in value):
            raise ValueError("amount must be non-negative")


@attrs.frozen(kw_only=True, auto_detect=True)
class GeneralLedgerLineImpl(GeneralLedgerLine[Account, Currency]):
    date: pd.Timestamp
    values: Sequence[LedgerElement[Account, Currency]]


def generalledger_line_to_multiledger_line(
    line: GeneralLedgerLine[Account, Currency], is_debit: Callable[[Account], bool], /
) -> MultiLedgerLine[Account, Currency]:
    """
    Convert a GeneralLedgerLine to a MultiLedgerLine.

    Parameters
    ----------
    line : GeneralLedgerLine[Account, Currency]
        The GeneralLedgerLine to convert.
    is_debit : Callable[[Account], bool]
        Whether the account is a debit account.

    Returns
    -------
    MultiLedgerLine[Account, Currency]
        The converted MultiLedgerLine.

    """
    debit = []
    credit = []
    for el in line.values:
        if is_debit(el.account) == (el.amount > 0):
            debit.append(
                LedgerElementImpl(
                    account=el.account, amount=abs(el.amount), currency=el.currency
                )
            )
        else:
            credit.append(
                LedgerElementImpl(
                    account=el.account, amount=abs(el.amount), currency=el.currency
                )
            )
    return MultiLedgerLineImpl(date=line.date, debit=debit, credit=credit)


def multiledger_line_to_generalledger_line(
    line: MultiLedgerLine[Account, Currency], /
) -> GeneralLedgerLine[Account, Currency]:
    """
    Convert a MultiLedgerLine to a GeneralLedgerLine.

    Parameters
    ----------
    line : MultiLedgerLine[Account, Currency]
        The MultiLedgerLine to convert.

    Returns
    -------
    GeneralLedgerLine[Account, Currency]
        The converted GeneralLedgerLine.

    """
    return GeneralLedgerLineImpl(
        date=line.date,
        values=[*line.debit, *line.credit],
    )


def multiledger_line_to_ledger_line(
    line: MultiLedgerLine[Account, Currency], /
) -> Sequence[LedgerLine[Account | AccountSundry, Currency]]:
    """
    Convert a MultiLedgerLine to a list of LedgerLine.

    Parameters
    ----------
    line : MultiLedgerLine[Account, Currency]
        The MultiLedgerLine to convert.

    Returns
    -------
    Sequence[LedgerLine[Account | AccountSundry, Currency]]
        The converted LedgerLines.

    """
    if (
        len(line.debit) == len(line.credit) == 1
        and line.debit[0].amount == line.credit[0].amount
        and line.debit[0].currency == line.credit[0].currency
    ):
        return [
            LedgerLineImpl(
                date=line.date,
                amount=line.debit[0].amount,
                currency=line.debit[0].currency,
                debit_account=line.debit[0].account,
                credit_account=line.credit[0].account,
            )
        ]
    return [
        LedgerLineImpl(
            date=line.date,
            amount=el.amount,
            currency=el.currency,
            debit_account=el.account,
            credit_account=SUNDRY,
        )
        for el in line.debit
    ] + [
        LedgerLineImpl(
            date=line.date,
            amount=el.amount,
            currency=el.currency,
            debit_account=SUNDRY,
            credit_account=el.account,
        )
        for el in line.credit
    ]


def generalledger_to_multiledger(
    lines: Sequence[GeneralLedgerLine[Account, Currency]],
    is_debit: Callable[[Account], bool],
) -> Sequence[MultiLedgerLine[Account, Currency]]:
    """
    Convert a GeneralLedger to a MultiLedger.

    Parameters
    ----------
    lines : Sequence[GeneralLedgerLine[Account, Currency]]
        The GeneralLedger to convert.
    is_debit : Callable[[Account], bool]
        Whether the account is a debit account.

    Returns
    -------
    Sequence[MultiLedgerLine[Account, Currency]]
        The converted MultiLedger.

    """
    return [generalledger_line_to_multiledger_line(line, is_debit) for line in lines]


def multiledger_to_generalledger(
    lines: Sequence[MultiLedgerLine[Account, Currency]],
) -> Sequence[GeneralLedgerLine[Account, Currency]]:
    """
    Convert a MultiLedger to a GeneralLedger.

    Parameters
    ----------
    lines : Sequence[MultiLedgerLine[Account, Currency]]
        The MultiLedger to convert.

    Returns
    -------
    Sequence[GeneralLedgerLine[Account, Currency]]
        The converted GeneralLedger.

    """
    return [multiledger_line_to_generalledger_line(line) for line in lines]


def multiledger_to_ledger(
    lines: Sequence[MultiLedgerLine[Account, Currency]],
) -> Sequence[LedgerLine[Account | AccountSundry, Currency]]:
    """
    Convert a MultiLedger to a Ledger.

    Parameters
    ----------
    lines : Sequence[MultiLedgerLine[Account, Currency]]
        The MultiLedger to convert.

    Returns
    -------
    Sequence[LedgerLine[Account | AccountSundry, Currency]]
        The converted Ledger.

    """
    return list(chain(*[multiledger_line_to_ledger_line(line) for line in lines]))



from collections.abc import Sequence
from pathlib import Path
from typing import Any

from .._ledger import GeneralLedgerLineImpl, LedgerElementImpl
from ._io import read_simple_csvs


def ledger_from_expenses(
    path: Path,
) -> Sequence[GeneralLedgerLineImpl[Any, Any]]:
    """
    Generate ledger from expenses.

    The CSV files are assumed to have columns
    ["勘定科目"].
    The relative path of the CSV file would be used as "取引先".

    Parameters
    ----------
    path : Path
        The path to the directory containing CSV files.

    Returns
    -------
    Sequence[GeneralLedgerLineImpl[Any, Any]]
        The ledger lines.

    """
    df = read_simple_csvs(path / "expenses")
    if df.empty:
        return []
    df["取引先"] = df["path"]
    res: list[GeneralLedgerLineImpl[Any, Any]] = []
    for date, row in df.iterrows():
        res.append(
            GeneralLedgerLineImpl(
                date=date,
                values=[
                    LedgerElementImpl(
                        account="事業主借", amount=row["金額"], currency=row["通貨"]
                    ),
                    LedgerElementImpl(
                        account=row["勘定科目"],
                        amount=row["金額"],
                        currency=row["通貨"],
                    ),
                ],
            )
        )
    return res



from collections.abc import Sequence
from decimal import ROUND_DOWN, Decimal, localcontext
from pathlib import Path
from typing import Any

import networkx as nx
import pandas as pd
from account_codes_jp import get_node_from_label

from .._ledger import GeneralLedgerLineImpl, LedgerElementImpl
from ._io import read_simple_csvs


def withholding_tax(amount: Decimal) -> Decimal:
    """
    Withholding tax calculation for most 源泉徴収が必要な報酬・料金等.

    Parameters
    ----------
    amount : Decimal
        The raw amount.

    Returns
    -------
    Decimal
        The withholding tax amount.

    References
    ----------
    https://www.nta.go.jp/taxes/shiraberu/taxanswer/gensen/2792.htm

    """
    with localcontext() as ctx:
        ctx.rounding = ROUND_DOWN
        if amount > 1000000:
            return round(
                Decimal(
                    1000000 * Decimal("0.1021") + (amount - 1000000) * Decimal("0.2042")
                ),
                0,
            )
        else:
            return round(amount * Decimal("0.1021"), 0)


def ledger_from_sales(
    path: Path,
    G: nx.DiGraph | None = None,
) -> Sequence[GeneralLedgerLineImpl[Any, Any]]:
    """
    Generate ledger from sales.

    The CSV files are assumed to have columns
    ["発生日", "金額", "振込日", "源泉徴収", "手数料"].
    If "源泉徴収" is True, the amount would be assumed by `withholding_tax()`.
    If "源泉徴収" if False or NaN, the amount would be assumed as 0.
    If "源泉徴収" is numeric, the amount would be assumed as 源泉徴収額.
    The relative path of the CSV file would be used as "取引先".

    Parameters
    ----------
    path : Path
        The path to the directory containing CSV files.
    G : nx.DiGraph | None
        The graph of accounts, by default None.
        If provided, each "取引先" would be added as a child node of "売上".

    Returns
    -------
    Sequence[GneralLedgerLineImpl[Any, Any]]
        The ledger lines.

    Raises
    ------
    ValueError
        If the transaction date is later than the transfer date.
    ValueError
        If withholding tax is included in transactions with different currencies.

    """
    df = read_simple_csvs(path / "sales")
    if df.empty:
        return []
    df["取引先"] = df["path"].str.replace(".csv", "")
    df["手数料"] = df["手数料"].apply(
        lambda x: Decimal(x) if pd.notna(x) else Decimal(0)
    )
    df["源泉徴収"] = df["源泉徴収"].replace(
        {"True": True, "False": False, "true": True, "false": False}
    )
    df.fillna({"源泉徴収": Decimal(0)}, inplace=True)

    if G is not None:
        for ca in ["売上", "仮払税金"]:
            parent_node = get_node_from_label(
                G, ca, lambda x: not G.nodes[x]["abstract"]
            )
            parent_node_attrs = G.nodes[parent_node]
            for t in df["取引先"].unique():
                t_attrs = {**parent_node_attrs, "label": f"{ca}({t})"}
                t_id = f"{ca}({t})"
                G.add_node(t_id, **t_attrs)
                G.add_edge(parent_node, t_id)

    ledger_lines: list[GeneralLedgerLineImpl[Any, Any]] = []
    for date, row in df.iterrows():
        ledger_lines.append(
            GeneralLedgerLineImpl(
                date=date,
                values=[
                    LedgerElementImpl(
                        account="売掛金", amount=row["金額"], currency=row["通貨"]
                    ),
                    LedgerElementImpl(
                        account="売上" if G is None else f"売上({row['取引先']})",
                        amount=row["金額"],
                        currency=row["通貨"],
                    ),
                ],
            )
        )
    for (t, date, currency), df_ in df.groupby(["取引先", "振込日", "通貨"]):
        fees = Decimal(df_["手数料"].sum())
        receivable = Decimal(df_["金額"].sum())
        recievable_without_fees = receivable - fees
        if currency == "":
            withholding = withholding_tax(
                df_.loc[df_["源泉徴収"] == True, "金額"].sum()
            )
            values = [
                LedgerElementImpl(
                    account="事業主貸",
                    amount=recievable_without_fees - withholding,
                    currency=currency,
                )
            ]
            if withholding > 0:
                values.append(
                    LedgerElementImpl(
                        account=f"仮払税金({t})", amount=withholding, currency=currency
                    )
                )
        else:
            if (df_["源泉徴収"] == True).any():
                raise ValueError("通貨が異なる取引に源泉徴収が含まれています。")
            values = [
                LedgerElementImpl(
                    account="事業主貸",
                    amount=recievable_without_fees,
                    currency=currency,
                )
            ]
        if fees > 0:
            values.append(
                LedgerElementImpl(account="支払手数料", amount=fees, currency=currency)
            )
        ledger_lines.append(
            GeneralLedgerLineImpl(
                date=date,
                values=[
                    *values,
                    LedgerElementImpl(
                        account="売掛金", amount=-receivable, currency=currency
                    ),
                ],
            )
        )
    if (df["発生日"] > df["振込日"]).any():
        raise ValueError("発生日が振込日より後の取引があります。")
    return ledger_lines
